# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

import os
import sys
from argparse import ArgumentParser
import logging

from twisted.application import service
from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList, inlineCallbacks
from twisted.words.protocols.jabber.jid import JID
from twisted.python import log  # used for xml logging

from wokkel.client import XMPPClient

# endroid base layer
from endroid.rosterhandler import RosterHandler
from endroid.wokkelhandler import WokkelHandler
# top layer
from endroid.usermanagement import UserManagement
from endroid.messagehandler import MessageHandler
# utilities
from endroid.confparser import Parser
from endroid.database import Database
from endroid.manhole import start_manhole


__version__ = (1, 2)

LOGGING_FORMAT = '%(asctime)-8s %(levelname)-8s %(message)s'
LOGGING_DATE_FORMAT = '%H:%M:%S'
# LOGGING_FORMAT = '%(levelname)-5s: %(message)s'


class Endroid(object):
    def __init__(self, conffile, logtraffic=False):
        self.application = service.Application("EnDroid")

        self.conf = Parser(conffile)

        self.jid = self.conf.get("setup", "jid")
        logging.info("Found JID: " + self.jid)

        self.secret = self.conf.get("setup", "password")
        logging.info("Found Secret: **********")

        rooms = self.conf.get("setup", "rooms", default=[])
        for room in rooms:
            logging.info("Found Room to Join: " + room)

        groups = self.conf.get("setup", "groups", default=['all'])
        for group in groups:
            logging.info("Found Group: " + group)

        dbfile = self.conf.get("database", "dbfile")
        logging.info("Using " + dbfile + " as database file")
        Database.setFile(dbfile)

        logfile = self.conf.get("setup", "logfile",
                                default='~/.endroid/endroid.log')
        logging.info("Using " + logfile + " as xml log file")
        log.startLogging(open(os.path.expanduser(logfile), "w+"))

        self.client = XMPPClient(JID(self.jid), self.secret)
        logging.info("Setting traffic logging to " + str(logtraffic))
        self.client.logTraffic = logtraffic

        self.client.setServiceParent(self.application)

        self.rosterhandler = RosterHandler()
        self.rosterhandler.setHandlerParent(self.client)

        self.wokkelhandler = WokkelHandler()
        self.wokkelhandler.setHandlerParent(self.client)

        self.usermanagement = UserManagement(self.wokkelhandler, self.rosterhandler, self.conf)
        self.messagehandler = MessageHandler(self.wokkelhandler, self.usermanagement)

        # Fire off our startup flow (once the reactor is running)
        reactor.callWhenRunning(self.startup_flow)

    @inlineCallbacks
    def startup_flow(self):
        # Start the client!
        self.client.startService()

        # wait for the wokkelhandler and rosterhandler to connect
        whd = Deferred()
        rhd = Deferred()
        self.wokkelhandler.set_connected_handler(whd)
        self.rosterhandler.set_connected_handler(rhd)
        yield DeferredList([whd, rhd])

    def run(self):
        reactor.run()


def main():
    parser = ArgumentParser(prog="endroid",
                            epilog="I'm a robot. I'm not a refrigerator.",
                            description="EnDroid: Extensible XMPP Bot")
    parser.add_argument("conffile", nargs='?', default="",
                        help="Configuration file to use.")
    parser.add_argument("-l", "--level", type=int, default=logging.INFO,
                        help="Logging level. Lower is more verbose.")
    parser.add_argument("-L", "--logfile", default=None,
                        help="File for logging output.")
    parser.add_argument("-t", "--logtraffic", action='store_true',
                        help="Additionally log all traffic.")
    parser.add_argument("-m", "--manhole", nargs='+', default=None,
                        help="Login name, password and port for ssh access")
    args = parser.parse_args()

    if args.logfile:
        logging.basicConfig(filename=args.logfile, level=args.level,
                            format=LOGGING_FORMAT, datefmt=LOGGING_DATE_FORMAT)
    else:
        logging.basicConfig(level=args.level, format=LOGGING_FORMAT, datefmt=LOGGING_DATE_FORMAT)

    cmd = args.conffile
    env = os.environ.get("ENDROID_CONF", "")
    usr = os.path.expanduser(os.path.join("~", ".endroid", "endroid.conf"))
    gbl = "/etc/endroid/endroid.conf"

    try:
        conffile = (p for p in (cmd, env, usr, gbl) if os.path.exists(p)).next()
    except StopIteration:
        logging.error("EnDroid requires a configuration file.")
        sys.exit(1)

    logging.info("EnDroid starting up with conffile {0}".format(conffile))

    droid = Endroid(conffile, logtraffic=args.logtraffic)

    if args.manhole:
        manhole_dict = dict([('droid', droid)] + globals().items())
        start_manhole(droid, manhole_dict, args.manhole)
    else:
        droid.run()

if __name__ == "__main__":
    main()
