# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

import logging
from twisted.words.protocols.jabber.jid import JID


class Handler(object):
    __slots__ = ("name", "priority", "callback")
    def __init__(self, priority, callback):
        self.name = callback.__name__
        self.priority = priority
        self.callback = callback

    def __str__(self):
        return "{}: {}".format(self.priority, self.name)


class MessageHandler(object):
    """An abstraction of XMPP's message protocols."""

    PRIORITY_NORMAL = 0
    PRIORITY_URGENT = -1
    PRIORITY_BULK = 1

    def __init__(self, wh, um):
        self.wh = wh
        self.um = um
        # wh translates messages and gives them to us, needs to know who we are
        self.wh.set_message_handler(self)
        self._handlers = {}

    def register_callback(self, name, typ, cat, callback,
                          including_self=False, priority=PRIORITY_NORMAL):
        """
        Register a function to be called on receipt of a message of type 
        'typ' (muc/chat), category 'cat' (recv, send, unhandled, *_self, *_filter)
        sent from user or room 'name'.

        """
        # self._handlers is a dictionary of form:
        # { type : { category : { room/groupname : [Handler objects]}}}
        typhndlrs = self._handlers.setdefault(typ, {})
        cathndlrs = typhndlrs.setdefault(cat, {})
        handlers = cathndlrs.setdefault(name, [])
        handlers.append(Handler(priority, callback))
        handlers.sort(key=lambda h: h.priority)

        # this callback be called when we get messages sent by ourself
        if including_self:
            self.register_callback(name, typ, cat + "_self", callback,
                                   priority=priority)

    def get_handlers(self, typ, cat, name):
        dct = self._handlers.get(typ, {}).get(cat, {})
        if typ == 'chat':  # we need to lookup name's groups
            # we may have either a full jid or just a userhost,
            # groups are referenced by userhost
            name = self.um.get_userhost(name)
            handlers = []
            for name in self.um.get_groups(name):
                handlers.extend(dct.get(name, []))
            handlers.sort(key=lambda h: h.priority)
            return handlers
        else:  # we are in a room so only one set of handlers to read
            return dct.get(name, [])

    def get_filters(self, typ, cat, name):
        return self.get_handlers(typ, cat+"_filter", name)

    def do_callback(self, cat, msg, failback):
        if msg.place == "muc":
            # get the handlers active in the room - note that these are already
            # sorted (sorting is done in the register_callback method)
            handlers = self.get_handlers(msg.place, cat, msg.recipient)
            filters = self.get_filters(msg.place, cat, msg.recipient)
        else: 
            # combine the handlers from each group the user is registered with
            # note that if the same plugin is registered for more than one of
            # the user's groups, the plugin's instance in each group will be
            # called
            handlers = self.get_handlers(msg.place, cat, msg.sender)
            filters = self.get_filters(msg.place, cat, msg.sender)
   
        log_list = []
        if handlers and all(f.callback(msg) for f in filters):
            msg.set_unhandled_cb(failback)

            for i in handlers:
                msg.inc_handlers()

            log_list.append("Did {} {} handlers (priority: cb):".format(len(handlers), cat))
            for handler in handlers:
                try:
                    handler.callback(msg)
                    log_list.append(str(handler))
                except Exception as e:
                    log_list.append("Exception in {}:\n{}".format(handler.name, e))
                    msg.dec_handlers()
                    raise
        else:
            failback(msg)
        if log_list:
            logging.info("Finished plugin callback: {}".format(
                         "\n\t".join(log_list)))
        else:
            logging.info("Finished plugin callback - no plugins called.")

    def _unhandled_muc(self, msg):
        self.do_callback("unhandled", msg, lambda m: None)

    def _unhandled_self_muc(self, msg):
        self.do_callback("unhandled_self", msg, lambda m: None)

    # Do normal (recv) callbacks on msg. If no callbacks handle the message
    # then call unhandled callbacks (msg's failback is set self._unhandled_...
    # by the last argument to do_callback).
    def receive_muc(self, msg):
        self.do_callback("recv", msg, self._unhandled_muc)

    def receive_self_muc(self, msg):
        self.do_callback("recv_self", msg, self._unhandled_self_muc)

    def _unhandled_chat(self, msg):
        self.do_callback("unhandled", msg, lambda m: None)

    def _unhandled_self_chat(self, msg):
        self.do_callback("unhandled_self", msg, lambda m: None)

    def receive_chat(self, msg):
        self.do_callback("recv", msg, self._unhandled_chat)

    def receive_self_chat(self, msg):
        self.do_callback("recv_self", msg, self._unhandled_self_chat)

    def send_muc(self, room, body, source=None, priority=PRIORITY_NORMAL):
        """
        Send muc message to room.

        The message will be run through any registered filters before it is sent.

        """
        msg = Message('muc', source, body, self, recipient=room)
        # when sending messages we check the filters registered with the
        # _recipient_. Cf. when we receive messages we check filters registered
        # with the _sender_.
        filters = self.get_filters('muc', 'send', msg.recipient)

        if all(f.callback(msg) for f in filters):
            logging.info("Sending message to {}".format(room))
            self.wh.groupChat(JID(room), body)
        else:
            logging.info("Filtered out message to {}".format(room))

    def send_chat(self, user, body, source=None, priority=PRIORITY_NORMAL):
        """
        Send chat message to person with address user.

        The message will be run through any registered filters before it is sent.

        """
        msg = Message('chat', source, body, self, recipient=user)
        filters = self.get_filters('chat', 'send', msg.recipient)

        if all(f.callback(msg) for f in filters):
            logging.info("Sending message to {}".format(user))
            self.wh.chat(JID(user), body)
        else:
            logging.info("Filtered out message to {0}".format(user))


class Message(object):
    def __init__(self, place, sender, body, messagehandler, recipient, handlers=0,
                 priority=MessageHandler.PRIORITY_NORMAL):
        self.place = place

        # sender_full is a string representing the full jid (including resource)
        # of the message's sender. Used in reply methods so that if a user is
        # logged in on several resources, the reply will be sent to the right one
        self.sender_full = sender
        # a string represeting the userhost of the message's sender. Used to
        # lookup resource-independant user properties eg their registered rooms
        self.sender = messagehandler.um.get_userhost(sender)
        self.body = body
        self.recipient = recipient

        # a count of plugins which will try to process this message
        self.__handlers = handlers
        self.messagehandler = messagehandler
        self.priority = priority

    def send(self):
        if self.place == "chat":
            self.messagehandler.send_chat(self.recipient, self.body, self.sender)
        elif self.place == "muc":
            self.messagehandler.send_muc(self.recipient, self.body, self.sender)

    def reply(self, body):
        if self.place == "chat":
            self.messagehandler.send_chat(self.sender_full, body, self.recipient)
        elif self.place == "muc":
            # we send to the room (the recipient), not the message's sender
            self.messagehandler.send_muc(self.recipient, body, self.recipient)

    def reply_to_sender(self, body):
        self.messagehandler.send_chat(self.sender_full, body, self.recipient)

    def inc_handlers(self):
        self.__handlers += 1

    def dec_handlers(self):
        self.__handlers -= 1
        self.do_unhandled()

    def unhandled(self, *args):
        """
        Notify the message that the caller hasn't handled it. This should only
        be called by plugins that have registered as a handler (and thus have
        incremented the handler count for this message).

        This method takes arbitrary arguments so it can be used as deferred
        callback or errback.

        """
        self.dec_handlers()

    def do_unhandled(self):
        if self.__handlers == 0 and hasattr(self, 'unHandledCallback'):
            self.unHandledCallback(self)

    def set_unhandled_cb(self, cb):
        self.unHandledCallback = cb

