# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

import sys
import logging
from endroid.cron import Cron

msg_filter_doc = ("Register a {0} filter for {1} messages.\n"
    "Filter takes endroid.messagehandler.Message and returns bool. If its\n"
    "return value evaluates to False, the message is dropped.\n"
    "Priority specifies the order of calling - lower numbers = called earlier.\n")

msg_send_doc = ("Send a {0} message to {1}, with given text and priority.\n"
    "rjid is a string representing the JID of the room to send to.\n"
    "text is a string (or unicode) of the message contents.\n"
    "priority one of the PRIORITY_* constants from endroid.messagehandler.\n"
    "(Use with care, especially PRIORITY_URGENT, which will usually bypass any\n"
    "ratelimiting or other protections put in place.)")

msg_cb_doc = ("Register a callback for {0} messages.\n"
    "Callback takes endroid.messagehandler.Message and may alter it arbitrarily.\n"
    "Inc_self specifies whether to do the callback if EnDroid created the message.\n"
    "Priority specifies the order of calling - lower numbers = called earlier.\n")


class PluginMeta(type):
    """
    Metaclass for Plugins that automatically registers any Plugin subclasses
    against their module name. This allows the PluginManager to automatically
    find the right class (meaning plugins don't need to register themselves,
    nor provide some well-known function like 'get_plugin').

    The registration key being the module is consistent with the expectation
    that the plugin name in the config file is a python module.

    This metaclass also ensures the Plugin has certain fields, and converts
    them to properties if required.
    """
    registry = {}

    def __new__(meta, name, bases, dict):
        if 'name' not in dict:
            dict['name'] = name.lower()
        for prop in ('dependencies', 'preferences'):
            if callable(dict.get(prop, None)):
                dict[prop] = property(dict[prop])

        # For backwards compatibility
        if 'enInit' in dict:
            dict['endroid_init'] = dict['enInit']
            del dict['enInit']
        return type.__new__(meta, name, bases, dict)

    def __init__(cls, name, bases, dict):
        type.__init__(cls, name, bases, dict)
        cls.modname = cls.__module__
        PluginMeta.registry[cls.__module__] = cls


class Plugin(object):
    """
    Parent class of all plugin objects within EnDroid. Plugins must subclass
    this type and it also represents the entry point for the plugin into the
    rest of EnDroid.

    """
    __metaclass__ = PluginMeta

    def _pluginInit(self, pm, conf):
        self._pm = pm

        self.messagehandler = pm.messagehandler
        self.usermanagement = pm.usermanagement

        self.place = pm.place
        self.place_name = pm.name
        self.vars = conf

    def _register(self, *args, **kwargs):
        return self.messagehandler.register_callback(self._pm.name, *args, **kwargs)

    # Registration methods
    def register_muc_callback(self, callback, inc_self=False, priority=0):
        if self._pm.place != "room":
            return
        self._register("muc", "recv", callback, inc_self, priority)

    def register_chat_callback(self, callback, inc_self=False, priority=0):
        if self._pm.place != "group":
            return
        self._register("chat", "recv", callback, inc_self, priority)


    def register_unhandled_muc_callback(self, callback, inc_self=False, priority=0):
        if self._pm.place != "room":
            return
        self._register("muc", "unhandled", callback, inc_self, priority)

    def register_unhandled_chat_callback(self, callback, inc_self=False, priority=0):
        if self._pm.place != "group":
            return
        self._register("chat", "unhandled", callback, inc_self, priority)

    register_muc_callback.__doc__ = msg_cb_doc.format("muc")
    register_chat_callback.__doc__ = msg_cb_doc.format("chat")
    register_unhandled_muc_callback.__doc__ = msg_cb_doc.format("unhandled muc")
    register_unhandled_chat_callback.__doc__ = msg_cb_doc.format("unhandled chat")

    def register_muc_filter(self, callback, inc_self=False, priority=0):
        if self._pm.place != "room":
            return
        self._register("muc", "recv_filter", callback, inc_self, priority)

    def register_chat_filter(self, callback, inc_self=False, priority=0):
        if self._pm.place != "group":
            return
        self._register("chat", "recv_filter", callback, inc_self, priority)

    def register_muc_send_filter(self, callback, inc_self=False, priority=0):
        if self._pm.place != "room":
            return
        self._register("muc", "send_filter", callback, inc_self, priority)

    def register_chat_send_filter(self, callback, inc_self=False, priority=0):
        if self._pm.place != "group":
            return
        self._register("chat", "send_filter", callback, inc_self, priority)

    register_muc_filter.__doc__ = msg_filter_doc.format("receive", "muc")
    register_chat_filter.__doc__ = msg_filter_doc.format("receive", "chat")
    register_muc_send_filter.__doc__ = msg_filter_doc.format("send", "muc")
    register_chat_send_filter.__doc__ = msg_filter_doc.format("send", "chat")

    # Plugin access methods
    def get(self, plugin_name):
        """Return a plugin-like object from the plugin module plugin_name."""
        return self._pm.get(plugin_name)

    def get_dependencies(self):
        """
        Return an iterable of plugins this plugin depends on.

        This includes indirect dependencies i.e. the dependencies of plugins this
        plugin depends on and so on.

        """
        return (self.get(dependency) for dependency in self.dependencies)

    def get_preferences(self):
        """
        Return an iterable of plugins this plugin prefers.

        This includes indirect preferences i.e. the preferences of plugins this
        plugin prefers and so on.

        """
        return (self.get(preference) for preference in self.preferences)

    def list_plugins(self):
        """Return a list of all plugins loaded in the plugin's environment."""
        return self._pm.get_plugins()

    def pluginLoaded(self, modname):
        """Check if modname is loaded in the plugin's environment (bool)."""
        return self._pm.hasloaded(modname)

    def pluginCall(self, modname, func, *args, **kwargs):
        """Directly call a method on plugin modname."""
        return getattr(self.get(modname), func)(*args, **kwargs)

    def endroid_init(self):
        pass

    @property
    def cron(self):
        return Cron().get()
    

    dependencies = ()
    preferences = ()


class PluginProxy(object):
    def __init__(self, modname):
        __import__(modname)
        # dictionary mapping module names to module objects
        m = sys.modules[modname]
        # In loading a plugin, we first look for a get_plugin() function,
        # then a function with the same name as the module, and finally we
        # just check the automatic Plugin registry for a Plugin defined in
        # that module.
        if hasattr(m, 'get_plugin'):
            self.module = getattr(m, 'get_plugin')()
        else:
            self.module = PluginMeta.registry[modname]()

    def __getattr__(self, key):
        if hasattr(self.module, key):
            return getattr(self.module, key)
        else:
            return self

    def hasattr(self, key):
        return hasattr(self.module, key)

    def __call__(self, *args, **kwargs):
        return self

    def __nonzero__(self):
        return False


class ModuleNotLoadedError(Exception):
    def __init__(self, value):
        super(ModuleNotLoadedError, self).__init__(value)
        self.value = value

    def __str__(self):
        return repr(self.value)


class PluginInitError(Exception):
    pass


class PluginManager(object):
    def __init__(self, messagehandler, usermanagement, place, name, config):
        self.messagehandler = messagehandler
        self.usermanagement = usermanagement

        self.place = place
        self.name = name

        # this is a dictionary of plugin module names to pluginproxy objects
        self._loaded = {}
        # a dict of modnames : plugin configs
        self._plugins = {}
        # module name to bool dictionary
        self.initialised = {}

        self.read_config(config)
        self.loadPlugins()
        self.initPlugins()

    def read_config(self, conf):
        def get_data(modname):
            # return a tuple of (modname, modname's config)
            return modname, conf.get(self.place, self.name, "plugin", modname, default={})

        plugins = conf.get(self.place, self.name, "plugins")
        self._plugins = dict(map(get_data, plugins))

    def get_plugins(self):
        return self._plugins.keys()

    def load(self, modname):
        # loads the plugin module and adds a key to self._loaded
        logging.debug("\tLoading Plugin: " + modname)
        try:
            p = PluginProxy(modname)
            p._pluginInit(self, self._plugins[modname])

            self._loaded[modname] = p
        except ImportError as i:
            logging.error(i)
            logging.error("**Could Not Import Plugin \"" + modname
                          + "\". Check That It Exists In Your PYTHONPATH.")

    def loadPlugins(self):
        logging.info("Loading Plugins for {0}".format(self.name))
        plugins = self.get_plugins()
        for p in plugins:
            self.load(p)

    def hasloaded(self, name):
        return name in self._loaded

    def init(self, modname):
        if modname in self.initialised:
            logging.debug("\t{0} Already Initialised".format(modname))
            return True
        if not self.hasloaded(modname):
            logging.error("\t**Cannot Initialise Plugin \"" + modname + "\", "
                          "It Has Not Been Imported")
            return False
        logging.debug("\tInitialising Plugin: " + modname)

        # deal with dependencies and preferences
        plugin = self.get(modname)
        for mod_dep_name in plugin.dependencies:
            logging.debug("\t{0} Depends On {1}".format(modname, mod_dep_name))
            if not self.init(mod_dep_name):
                # can't possibly initialise us so remove us from self._loaded
                logging.error("\t**No \"" + mod_dep_name + "\". Unloading " + modname)
                self._loaded.pop(modname)
                return False

        for mod_pref_name in plugin.preferences:
            logging.debug("\t{0} Prefers {1}".format(modname, mod_pref_name))
            if not self.init(mod_pref_name):
                logging.error("\t**Could Not Load " + mod_pref_name)
        # attempt to initialise the plugin
        try:
            plugin.endroid_init()
            self.initialised[modname] = True
            logging.info("\tInitialised Plugin: " + modname)
        except PluginInitError:
            logging.error("\t**Error initializing \"" + modname + "\".  See "
                          "log for details.")
            return False
        return True

    def initPlugins(self):
        logging.info("Initialising Plugins for {0}".format(self.name))
        plugins = self.get_plugins()
        for p in plugins:
            self.init(p)

    def get(self, name):
        if not name in self._loaded:
            raise ModuleNotLoadedError(name)
        return self._loaded[name]
