from endroid.plugins.command import CommandPlugin
import shlex

def parse_string(string, options=None):
    """
    Parse a shell command like string into an args tuple and a kwargs dict.

    Options is an iterable of string tuples, each tuple representing a keyword
    followed by its synonyms.

    Words in string will be appended to the args tuple until a keyword is
    reached, at which point they will be appended to a list in kwargs.

    Eg parse_string("a b c -u 1 2 3 -r 5 6", [("-u",), ("room", "-r")])
    will return: ('a', 'b', 'c'), {'-u': ['1','2','3'], 'room': ['5','6']}

    """
    options = options or []
    aliases = {}

    keys = []
    # build the kwargs dict with all the keywords and an aliases dict for synonyms
    for option in options:
        if isinstance(option, (list, tuple)):
            main = option[0]
            keys.append(main)
            for alias in option[1:]:
                aliases[alias] = main
        elif isinstance(option, (str, unicode)):
            keys.append(option)

    args = []
    kwargs = {}
    current = None
    # parse the string - first split into shell 'words'
    parts = shlex.split(string)
    # then add to args or the kwargs dictionary as appropriate
    for part in parts:
        # if it's a synonym get the main command, else leave it
        part = aliases.get(part, part)
        if part in keys:
            # we have come to a keyword argument - create its list
            kwargs[part] = []
            # keep track of where we are sending non-keyword words to
            current = kwargs[part]
        elif current is not None:
            # we are adding words to a keyword's list
            current.append(part)
        else:
            # no keywords has been found yet - we are still in args
            args.append(part)

    return args, kwargs

def replace(l, search, replace):
    return [replace if item == search else item for item in l]


class Inviter(CommandPlugin):
    help = "Invite users to rooms"
    name = "invite"
    PARSE_OPTIONS = ("to", "into")

    def cmd_invite_me(self, msg, arg):
        """
        Invite user to the rooms listed in args, or to all their rooms 
        if args is empty.

        """
        args, kwargs = parse_string(arg, self.PARSE_OPTIONS)
        users = [msg.sender]
        rooms = set(args + kwargs.get("to", [])) or ["all"]

        results = self._do_invites(users, rooms)
        msg.reply(results)
    cmd_invite_me.helphint = "{to|into}? <room>+"

    def cmd_invite_all(self, msg, arg):
        """Invite all of a room's registered users to the room."""
        users = self.usermanagement.get_available_users(self.place_name)
        rooms = [self.place_name]

        results = self._do_invites(users, rooms)
        msg.reply(results)
    cmd_invite_all.muc_only = True
    cmd_invite_all.helphint = "<reason>"

    def cmd_invite_users(self, msg, arg):
        """Invite a list of users to a list of rooms."""
        args, kwargs = parse_string(arg, self.PARSE_OPTIONS)
        users = replace(args, "me", msg.sender)
        rooms = kwargs.get("to", [])

        results = self._do_invites(users, rooms)
        msg.reply(results)
    cmd_invite_users.helphint = "<user>+ {to|into} <room>+"

    def _do_invites(self, users, rooms):
        if 'all' in users:
            if len(rooms) == 1:
                users = self.usermanagement.get_available_users(rooms[0])
            else:
                return "Can only invite 'all' users to a single room"
        if 'all' in rooms:
            if len(users) == 1:
                rooms = self.usermanagement.get_rooms(users[0])
                if not rooms:
                    return "User '{}' is not registered.".format(users[0])
            else:
                return "Can only invite a single user to 'all' rooms"

        results = []
        invitations = 0
        for room in rooms:
            for user in users:
                s, reason = self.usermanagement.invite(user, room)
                if not s:
                    results.append("{} to {} failed: {}".format(user, room, reason))
                else:
                    invitations += 1
        return "Sent {} invitations\n".format(invitations) + '\n'.join(results)
